#!/usr/bin/env python
# coding: utf-8

# In[1]:


# Global Variables for this script - PART 01
# -------------------
sen_installation_folder = r"C:\Program Files\SCIA\Engineer22.1"
sen_temp_folder = r"C:\Users\bjorn\ESA22.1\Temp"
this_app_log_folder = r"C:\Temp\OpenAPI_Logs"
safxlsx_fname = r"C:\Temp\AdmSaf.xlsx"


# In[2]:


# Import modules and create helper routines
# -------------------

#import sys
import os
import clr
import System
import System.IO
#import System.Collections.Generic


def import_dll_by_name(dll, *folder_tup):
    """ Adds a reference to a dll,
    look for the dll in the list of possible folders
    With:
     - dll = the name of the dll
     - *folder_tup = a tuple that contains a variable number of folders;
                     the dll will be searched
    """
    for dllpath in folder_tup:
        dllpathfile = os.path.join(dllpath, dll)
        if os.path.isfile(dllpathfile):
            try:
                clr.AddReference(dllpathfile)
                print(f"Add reference to:  {dllpathfile}")
                return
            except Exception as e:
                print(f"FileNotFoundException thrown for:  {dllpathfile}")
                print(e)
                
    
    print(f"ERROR: Failed to add reference to {dll}")
    for dllpath in folder_tup:
        print(f"  - file not found in folder: {dllpath}")
    raise FileNotFoundError
    return

def newGuid():
    """create and return a new System guid
    """
    return System.Guid.NewGuid()

def obj_set_attrs(obj, *attr_d_tup):
    """ set extra attributes for an object,
    and return the object
    With:
     - obj = an adm object
     - *attr_d_tup = a tuple that contains a variable number of dicts;
                     each dict has (one or more) key = attribute name (as a string) and value = corresponding attribute value
    """
    for attr_d in attr_d_tup:
        for attr, val in attr_d.items():
            setattr(obj, attr, val)
    return obj




# In[3]:


# Add references to the SCIA.OpenAPI Dlls
# -------------------

admrefs_1 = [ "UnitsNet.dll",
              "System.ValueTuple.dll", ]
 
admrefs_2 = [ "ModelExchanger.Excel.Conversion.Contracts.dll",
              "ModelExchanger.Excel.Conversion.Implementation.dll",
              "ModelExchanger.Excel.Integration.Bootstrapper.dll", ]

admrefs_3 = [ "AutoMapper.dll",
              "AutoMapper.Extensions.EnumMapping.dll", ]

admrefs_4 = [ "ModelExchanger.AnalysisDataModel.dll",
              "ModelExchanger.AnalysisDataModel.Implementation.dll",
              "ModelExchanger.AnalysisDataModel.Contracts.dll",
              "ModelExchanger.AnalysisDataModel.Integration.Bootstrapper.dll",
              "ModelExchanger.Shared.Implementation.dll",
              "ModelExchanger.Shared.Models.dll",
              "CSInfrastructure.dll",
              "FluentValidation.dll",
              "Newtonsoft.Json.dll",
              "SimpleInjector.dll", ]

admrefs_5 = [ "Infrastructure.Geometry.dll",
              "MathNet.Spatial.dll",
              "MathNet.Numerics.dll", ]


for dll in set(admrefs_1 + admrefs_2 + admrefs_3 + admrefs_4 + admrefs_5):
    import_dll_by_name(dll, os.path.join(sen_installation_folder, "OpenAPI_dll"))


from ModelExchanger.Excel import *
from ModelExchanger.Excel.Integration.Bootstrapper import *
from ModelExchanger.Excel.Conversion.Contracts import *

from ModelExchanger.AnalysisDataModel import *
from ModelExchanger.AnalysisDataModel.Integration.Bootstrapper import *
from ModelExchanger.AnalysisDataModel.Base import *
from ModelExchanger.AnalysisDataModel.Contracts import *
import ModelExchanger.AnalysisDataModel.Enums as MADM_Enum
from ModelExchanger.AnalysisDataModel.Enums import *
from ModelExchanger.AnalysisDataModel.Interfaces import *
from ModelExchanger.AnalysisDataModel.Libraries import *
from ModelExchanger.AnalysisDataModel.Loads import *
from ModelExchanger.AnalysisDataModel.Models import *
from ModelExchanger.AnalysisDataModel.Results import *
from ModelExchanger.AnalysisDataModel.StructuralElements import *
from ModelExchanger.AnalysisDataModel.StructuralReferences import *
from ModelExchanger.AnalysisDataModel.StructuralReferences.Points import *
from ModelExchanger.AnalysisDataModel.StructuralReferences.Curves import *
from ModelExchanger.AnalysisDataModel.Subtypes import *
from ModelExchanger.AnalysisDataModel.Subtypes.Load import *

from CSInfrastructure import *
from CSInfrastructure.IOC import *

from UnitsNet import *
from UnitsNet.Units import *


# In[4]:


#LINQ-alike functions in python
#see also:  https://www.c-sharpcorner.com/blogs/linq-in-python-for-c-sharp

def getAdmObj(_name, *, _type=None):
    """ helper routine to get an ADM object by its name, and (optionally) filtered by it's type
    With:
      - _name =  name of the ADM object
      - _type =  name of the object-type
    """
    if not isinstance(_name, str):
        #for more flexibility:
        #  when you input the object itself (in stead of the name of the object), then the object is returned
        return _name
    if _type is None:
        retval = next( (it for it in model if it.Name == _name), None)       #equivalent to FirstOrDefault
    else:
        retval = next( (it for it in model if it.Name == _name and _type in it.GetType().ToString()), None)
    return retval

def getAdmObj_lst(_type):
    """ helper routine to get a list of all ADM objects by their type
    With:
      - _type =  name of the object-type
    """
    retvals = [it for it in model if _type in it.GetType().ToString()]
    return retvals


# In[ ]:





# In[5]:


#def main():

model = AnalysisModel()

print("Creating model...")
admBootstrapper = AnalysisDataModelBootstrapper()
admScope = admBootstrapper.CreateThreadedScope()
admService = admScope.GetService[ IAnalysisModelService ]()


# In[6]:


# add ModelInformation
def addModelInformation():
    _model_information = ModelInformation(newGuid(), "GinoModelInfo")
    obj_set_attrs(_model_information, {"NationalCode": NationalCode.EC_Standard_EN, "SourceApplication": "pySciaApi", "SourceCompany": "GinoSoft", "Owner": "Gino Vanstraelen", "LastUpdate": System.DateTime.Now})
    admService.AddItemToModel(model, _model_information)
    return _model_information

addModelInformation()


# In[7]:


# add Materials
def add_StructuralMaterial(name):
    #determine material type
    _materialType = MaterialType.Steel
    if ( name.startswith("C") and (r"/" in name) ) : _materialType = MaterialType.Concrete
    if ( ("(EN" in name) and (name.endswith("338)") or name.endswith("14080")) ): _materialType = MaterialType.Timber
    #return the material
    _material = StructuralMaterial(newGuid(), name, _materialType, name)
    admService.AddItemToModel(model, _material)
    return _material

def get_StructuralMaterial(name):
    _material = getAdmObj(name, _type = "StructuralMaterial")
    if _material is None: _material = addStructuralMaterial(name)
    return _material

matlst = ["S 235", "S 275", "S 355", "S 450", "C25/30", "C30/37", "C35/45", "C40/50", "C45/55", "C50/60"]
for it in matlst:
    add_StructuralMaterial(it)


# In[8]:


# add Sections
def _add_Structural_Manufactured_CrossSection(name, typecode, material, formcode):
    _mat = get_StructuralMaterial(material)

    if   formcode == "fc1":
        _fc = FormCode.ISection
        _id = DescriptionId.EuropeanIBeam
    elif formcode == "fc2":
        _fc = FormCode.RectangularHollowSection
        _id = DescriptionId.RectangularHollowSection
    elif formcode == "fc3":
        _fc = FormCode.CircularHollowSection
        _id = DescriptionId.CircularHollowSection
    elif formcode == "fc4":
        _fc = FormCode.LSection
        _id = DescriptionId.LegAngle
    elif formcode == "fc5":
        _fc = FormCode.ChannelSection
        _id = DescriptionId.ChannelSection
    elif formcode == "fc6":
        _fc = FormCode.TSection
        _id = DescriptionId.TeeShape
    elif formcode == "fc7":
        _fc = FormCode.FullRectangularSection
        _id = DescriptionId.FlatBar
    elif formcode == "fc8":
        _fc = FormCode.FullCircularSection
        _id = DescriptionId.RoundBar
    elif formcode == "fc9":
        _fc = FormCode.TSectionFlipped
        _id = DescriptionId.TeeShape
    elif formcode == "fc10":
        _fc = FormCode.LSectionMirrored
        _id = DescriptionId.LegAngle
    else:
        typecode = "RD100"
        _fc = FormCode.FullCircularSection
        _id = DescriptionId.RoundBar
        _mat = getStructuralMaterial("S 235")
            
    _css = StructuralManufacturedCrossSection(newGuid(), name, _mat, typecode, _fc, _id)
    admService.AddItemToModel(model, _css)
    return _css

def _add_Structural_Parametric_CrossSection(name, typecode, material):
    _mat = get_StructuralMaterial(material)
    
    lst_Lengths = System.Collections.Generic.List[ Length ]()
    
    _tc = typecode[:4].lower()
    if len(typecode) > 4 and _tc == "rect":
        _id = ProfileLibraryId.Rectangle
        dims = typecode[4:].split("x")
        lst_Lengths.Add(Length(float(dims[0]), LengthUnit.Millimeter))
        lst_Lengths.Add(Length(float(dims[1]), LengthUnit.Millimeter))
    elif len(typecode) > 4 and _tc == "circ":
        _id = ProfileLibraryId.Circle
        diam = typecode[4:]
        lst_Lengths.Add(Length(float(diam), LengthUnit.Millimeter))
    else:
        _id = ProfileLibraryId.Circle
        lst_Lengths.Add(Length(100, LengthUnit.Millimeter))
    
    _css = StructuralParametricCrossSection(newGuid(), name, _mat, _id, lst_Lengths)
    admService.AddItemToModel(model, _css)
    return _css

def add_StructuralCrossSection(name, typecode, material, formcode = None):
    if formcode is None:
        _css = _add_Structural_Parametric_CrossSection(name, typecode, material)
    else:
        _css = _add_Structural_Manufactured_CrossSection(name, typecode, material, formcode)
    return _css

csslst = [ ("hea200", "HEA200", "S 235", "fc1"),
           ("ipe360", "IPE360", "S 235", "fc1"),
           ("c_300", "Circ300", "C30/37"),
           ("r_300x300", "Rect300x300", "C30/37"),]
for it in csslst:
    add_StructuralCrossSection(*it)


# In[9]:


# add Nodes
def add_StructuralPointConnection(name, X, Y, Z):
    _X = Length(X, LengthUnit.Meter)
    _Y = Length(Y, LengthUnit.Meter)
    _Z = Length(Z, LengthUnit.Meter)
    _node = StructuralPointConnection(newGuid(), name, _X, _Y, _Z)
    admService.AddItemToModel(model, _node)
    return _node

l, w, h = 10, 5, 3.5  #length (X), width (Y), height (Z)
nodelst = [ ("n1", 0, 0, 0), ("n2", l, 0, 0), ("n3", l, w, 0), ("n4", 0, w, 0),
            ("n5", 0, 0, h), ("n6", l, 0, h), ("n7", l, w, h), ("n8", 0, w, h) ]
for it in nodelst:
    add_StructuralPointConnection(*it)
        


# In[10]:


# helper functions to create a curve, or a list of curves
def adm_curve(curve_shape, curve_nodes):
    """ helper routine for creating a Curve
    With:
      - curve shape =  Line, Arc3P, ParaArc, Spline, Bezier, CircleCP, Circle3P
      - curve_nodes =  list of node names
    """
    b_nodes = System.Collections.Generic.List[ StructuralPointConnection ]()
    
    curve_shape = curve_shape.lower()
    if   curve_shape == ["line"]:
        _shape = CurveGeometricalShape.Line
        for it in curve_nodes[:2]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape in ["arc3p", "arc"]:
        _shape = CurveGeometricalShape.Arc3P
        for it in curve_nodes[:3]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape in ["paraarc", "para"]:
        _shape = CurveGeometricalShape.ParaArc
        for it in curve_nodes[:3]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape == "spline":
        _shape = CurveGeometricalShape.Spline
        for it in curve_nodes:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape == "bezier":
        _shape = CurveGeometricalShape.Bezier
        for it in curve_nodes:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape == "circlecp":
        _shape = CurveGeometricalShape.CircleCP
        for it in curve_nodes[:2]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    elif curve_shape == "circle3p":
        _shape = CurveGeometricalShape.Circle3P
        for it in curve_nodes[:3]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))
    else:
        _shape = CurveGeometricalShape.Line
        for it in curve_nodes[:2]:
            b_nodes.Add(getAdmObj(it, _type = "StructuralPointConnection"))

    b_curve = Curve[ StructuralPointConnection ](_shape, b_nodes)
    return b_curve

def adm_curves(*admcurves):
    """ helper routine for creating a List of Curves
    With:
     - admcurves = (unpacked) list of Curve
    """
    b_curves = System.Collections.Generic.List[ Curve[ StructuralPointConnection ] ]()
    for admcurve in admcurves:
        b_curves.Add(admcurve)
    return b_curves

def closed_polygon(*nodes):
    """ helper routine for creating a List of (Curve-shape, Curve-nodes) that form a closed polygon
        usefull for creating slabs, ... with simple shapes (e.g. rectangle, ...)
    With:
     - nodes = list of node-names
    """
    end_nodes = [it for it in nodes[1:]]
    end_nodes.append(nodes[0])
    retval = []
    for beg_node, end_node in zip(nodes, end_nodes):
        retval.append( ("Line", [beg_node, end_node]) )
    return retval
        
    
    
    


# In[ ]:





# In[11]:


# add Beam
def add_StructuralCurveMember(name, curve_shape, curve_nodes, css, layer = None, lcs_type = None, lcs_rotation = None, lcs_type_params = None, structural_type = None, fem_type = None, member_systemline = None , exc_y = None, exc_z = None):
    """ create a beam
    With:
      - name = unique name of beam
      - curve_shape = Line, Arc3P, ParaArc, Spline, Bezier
      - curve_nodes = a list with the of names of the nodes
      - css  = name of css
      - layer = name of layer
      - lcs_type = std, vectorY, vectorZ, pointY, Pointz
            - lcs_type = std > Standard definition of LCS with an angle -> lcs_type_params is the angle in degrees
            - lcs_type = vectorY or vectorZ > Definition of LCS through a vector for local Y or Z -> lcs_type_params is list with the X, Y, Z components of the vector
            - lcs_type = pointY or pointZ > Definition of LCS oriented towards a point3D for local Y or Z -> lcs_type_params is list with the X, Y, Z components of the point3D
      - structural_type = general, beam, column, gable column, secondary column, rafter, purlin, roof bracing, wall bracing, girt, truss chord, truss diagonal, plate rib, beam slab
      - fem_type = "standard", "axial force only"
      - member_systemline = centre, top, bottom, left, topleft, bottomleft, right, topright, bottomright
    """
 
    _b_curves = adm_curves(adm_curve(curve_shape, curve_nodes))
    _css = getAdmObj(css)
    beam = StructuralCurveMember(newGuid(), name, _b_curves, _css)
    
    #set layer
    _layer = "LayerADM" if layer is None else layer
    obj_set_attrs(beam, {"Layer": _layer})

    #set lcs_type
    """ ATTENTION: 
    due to BUG in v2.1.0:  REPLACE 'Standard' by 'VectorZ'
    """
    lcs_type = "vecz" if lcs_type is None else lcs_type
    """
    ======================================================
    """
    
    lcs_type = "std" if lcs_type is None else lcs_type
    if   lcs_type in ["std", "standard"]:
        print('->std<- ', end="")
        _curve_lcstype = CurveLCSType.Standard
    elif lcs_type in ["vecy", "vectory", "y"]:
        _curve_lcstype = CurveLCSType.VectorY
    elif lcs_type in ["vecz", "vectorz", "z"]:
        _curve_lcstype = CurveLCSType.VectorZ
    elif lcs_type in ["pty", "pointy"]:
        _curve_lcstype = CurveLCSType.PointY
    elif lcs_type in ["ptz", "pointz"]:
        _curve_lcstype = CurveLCSType.PointZ
    else:
        _curve_lcstype = CurveLCSType.Standard
    
    
    #set LCSAdjustement
    lcs_rotation = 0 if lcs_rotation is None else lcs_rotation
    lcs_type_params = [0, 0, 1] if lcs_type_params is None else lcs_type_params

    _lcs_adjustment = LCSAdjustment[ CurveLCSType ]()
    obj_set_attrs(_lcs_adjustment, {"LCS": _curve_lcstype, "Angle": Angle(lcs_rotation, AngleUnit.Degree),
                                    "X": Length(lcs_type_params[0], LengthUnit.Meter), "Y": Length(lcs_type_params[1], LengthUnit.Meter), "Z": Length(lcs_type_params[2], LengthUnit.Meter)})
    obj_set_attrs(beam, {"LCSAdjustment": _lcs_adjustment})
    
    
    #set structural_type
    structural_type = "general" if structural_type is None else structural_type
    if structural_type == "general":
        _type = Member1DType.General
    elif structural_type == "beam":
        _type = Member1DType.Beam
    elif structural_type == "column":
        _type = Member1DType.Column
    elif structural_type == "gable column":
        _type = Member1DType.GableColumn
    elif structural_type == "secondary column":
        _type = Member1DType.SecondaryColumn
    elif structural_type == "rafter":
        _type = Member1DType.Rafter
    elif structural_type == "purlin":
        _type = Member1DType.Purlin
    elif structural_type == "roof bracing":
        _type = Member1DType.RoofBracing
    elif structural_type == "wall bracing":
        _type = Member1DType.WallBracing
    elif structural_type == "girt":
        _type = Member1DType.Girt
    elif structural_type == "truss chord":
        _type = Member1DType.TrussChord
    elif structural_type == "truss diagonal":
        _type = Member1DType.TrussDiagonal
    elif structural_type == "plate rib":
        _type = Member1DType.PlateRib
    elif structural_type == "beam slab":
        _type = Member1DType.BeamSlab
    else:
        _type = Member1DType.General
    obj_set_attrs(beam, {"Type": FlexibleEnum[ Member1DType ](_type)})    
    
    #set fem_type
    fem_type = "standard" if fem_type is None else fem_type
    if   fem_type == "standard":
        _curve_behaviour = CurveBehaviour.Standard
    elif fem_type == "axial force only":
        _curve_behaviour = FlexibleEnum[ Member1DType ](CurveBehaviour.AxialForceOnly)
    else:
        _curve_behaviour = CurveBehaviour.Standard
    obj_set_attrs(beam, {"Behaviour": _curve_behaviour})
                  
    #set member_systemline
    member_systemline = "centre" if member_systemline is None else member_systemline
    if   member_systemline == "centre":
        _curveAlignment = CurveAlignment.Centre
    elif member_systemline == "top":
        _curveAlignment = CurveAlignment.Top
    elif member_systemline == "bottom":
        _curveAlignment = CurveAlignment.Bottom
    elif member_systemline == "left":
        _curveAlignment = CurveAlignment.Left
    elif member_systemline == "right":
        _curveAlignment = CurveAlignment.Right
    elif member_systemline == "topleft":
        _curveAlignment = CurveAlignment.TopLeft
    elif member_systemline == "topright":
        _curveAlignment = CurveAlignment.TopRight
    elif member_systemline == "bottomleft":
        _curveAlignment = CurveAlignment.BottomLeft
    elif member_systemline == "bottomright":
        _curveAlignment = CurveAlignment.BottomRight
    else:
        _curveAlignment = CurveAlignment.Centre
    obj_set_attrs(beam, {"SystemLine": _curveAlignment})
    
    #set exc_y and exc_z
    exc_y = 0 if exc_y is None else exc_y
    exc_z = 0 if exc_z is None else exc_z
    if not exc_y == 0: obj_set_attrs(beam, {"AnalysisEccentricityYBegin": Length(exc_y, LengthUnit.Millimeter), "AnalysisEccentricityYEnd": Length(exc_y, LengthUnit.Millimeter)})
    if not exc_z == 0: obj_set_attrs(beam, {"AnalysisEccentricityZBegin": Length(exc_z, LengthUnit.Millimeter), "AnalysisEccentricityZEnd": Length(exc_z, LengthUnit.Millimeter)})
    
    admService.AddItemToModel(model, beam)
    return beam


beamlst = [ ("b1", "Line", ["n1", "n5"], "hea200"),
            ("b2", "Line", ["n2", "n6"], "hea200"),
            ("b3", "Line", ["n3", "n7"], "hea200"),
            ("b4", "Line", ["n4", "n8"], "hea200"),
            ("b5", "Line", ["n5", "n6"], "ipe360"),
            ("b6", "Line", ["n6", "n7"], "ipe360"),
            ("b7", "Line", ["n7", "n8"], "ipe360"),
            ("b8", "Line", ["n8", "n5"], "ipe360"),]
for it in beamlst:
    add_StructuralCurveMember(*it)


# In[12]:


# add Plates
def add_StructuralSurfaceMember(name, material_name, thickness_mm, curve_lst):
    """ create a surface
    With:
     - name = name of slab
     - material name = name of material
     - tickness_mm = thickness of the slab, in mm
     - curve_lst = list of Curve
    """
    _mat = getAdmObj(material_name, _type="StructuralMaterial")
    
    _th = Length(thickness_mm, LengthUnit.Millimeter)
    
    _curves = adm_curves( *[adm_curve(it[0], it[1]) for it in curve_lst] )
    
    _slab = StructuralSurfaceMember(newGuid(), name, _curves, _mat, _th)
    admService.AddItemToModel(model, _slab)
    return _slab

def add_StructuralSurfaceMemberOpening(name, slab_name, curve_lst):
    """ create an opening in a flat surface
    With:
     - name = name of opening
     - slab_name = name of slab
     - curve_lst = list of Curve
    """
    _slab = getAdmObj(slab_name, _type="StructuralSurfaceMember")
    
    _curves = adm_curves( *[adm_curve(it[0], it[1]) for it in curve_lst])
    
    _o = StructuralSurfaceMemberOpening(newGuid(), name, _slab, _curves)
    admService.AddItemToModel(model, _o)
    return _o

def add_StructuralSurfaceMemberRegion(name, slab_name, curve_lst, material_name, thickness_mm):
    """ create a subregion in a flat surface
    With:
     - name = name of opening
     - slab_name = name of slab
     - curve_lst = list of Curve
     - material_name = name of material of the subregion
     - tickness_mm = thickness of the subregion, in mm
    """
    _slab = getAdmObj(slab_name, _type="StructuralSurfaceMember")
    _mat = getAdmObj(material_name, _type="StructuralMaterial")
    
    _curves = adm_curves( *[adm_curve(it[0], it[1]) for it in curve_lst])

    _r = StructuralSurfaceMemberRegion(newGuid(), name, _slab, _curves, _mat)
    
    _th = Length(thickness_mm, LengthUnit.Millimeter)
    
    obj_set_attrs(_r, {"Thickness": _th})
    
    admService.AddItemToModel(model, _r)
    return _r



nodelst = [ ("no1", 0.15*l, w, 1.5*h), ("no2", 0.2*l, w, 1.5*h), ("no3", 0.75*l, w, 1.5*h), ("no4", 0.8*l, w, 1.5*h),
            ("n9", 0, 0, 2*h), ("n10", l, 0, 2*h), ("n11", l, w, 2*h), ("n12", 0, w, 2*h), ("n13", l/2, w, 2*h + 0.4*l) ]
for it in nodelst:
    add_StructuralPointConnection(*it)

#slablst = [ ( "slab1", "C35/45", 300.0, [ ("Line", ["n5", "n6"]), ("Line", ["n6", "n7"]), ("Line", ["n7", "n8"]), ("Line", ["n8", "n5"]) ] ),
#            ( "wall1", "C35/45", 300.0, [ ("Line", ["n7", "n8"]), ("Line", ["n8", "n12"]), ("Arc", ["n12", "n13", "n11"]), ("Line", ["n11", "n7"]) ] )]

slablst = [ ( "slab1", "C35/45", 300.0, closed_polygon("n5", "n6", "n7", "n8") ),
            ( "wall1", "C35/45", 300.0, [ ("Line", ["n7", "n8"]), ("Line", ["n8", "n12"]), ("Arc", ["n12", "n13", "n11"]), ("Line", ["n11", "n7"]) ] )]

for it in slablst:
    add_StructuralSurfaceMember(*it)
                         
openlst = [ ("O1", "wall1", [ ("CircleCP", ["no2", "no1"]) ] ) ]
for it in openlst:
    add_StructuralSurfaceMemberOpening(*it)

regionlst = [ ("R1", "wall1", [ ("CircleCP", ["no4", "no3"]) ], "C35/45", 600.0 ) ]
for it in regionlst:
    add_StructuralSurfaceMemberRegion(*it)                         
                         
                         


# In[13]:


# add hinges1D
def h1d_tra_fix():
    """ helper routine for creating a Constraint for translation = rigid
    """
    it = Constraint[System.Nullable[ForcePerLength]]()
    obj_set_attrs(it, {"Type": ConstraintType.Rigid})
    return it
def h1d_tra_free():
    """ helper routine for creating a Constraint for translation = free
    """
    it = Constraint[System.Nullable[ForcePerLength]]()
    obj_set_attrs(it, {"Type": ConstraintType.Free})
    return it
def h1d_tra_flex(stiffness):
    """ helper routine for creating a Constraint for translation = flexible,  with stiffness in MN/m
    """
    it = Constraint[System.Nullable[ForcePerLength]]()
    obj_set_attrs(it, {"Type": ConstraintType.Flexible, "Stiffness": ForcePerLength(stiffness, ForcePerLengthUnit.MeganewtonPerMeter)})
    return it
def h1d_rot_fix():
    """ helper routine for creating a Constraint for rotation = rigid
    """
    it = Constraint[System.Nullable[RotationalStiffness]]()
    obj_set_attrs(it, {"Type": ConstraintType.Rigid})
    return it
def h1d_rot_free():
    """ helper routine for creating a Constraint for rotation = free
    """
    it = Constraint[System.Nullable[RotationalStiffness]]()
    obj_set_attrs(it, {"Type": ConstraintType.Free})
    return it
def h1d_rot_flex(stiffness):
    """ helper routine for creating a Constraint for rotation = flexible,  with stiffness in MNm/rad
    """
    it = Constraint[System.Nullable[RotationalStiffness]]()
    obj_set_attrs(it, {"Type": ConstraintType.Flexible, "Stiffness": RotationalStiffness(stiffness, RotationalStiffnessUnit.MeganewtonMeterPerRadian)})
    return it
def h1d_get_constraints(x_y_z_rx_ry_rz, stiffx, stiffy, stiffz, stiffrx, stiffry, stiffrz):
    """ helper routine that returns the 6 Constraints
    """
    _x  = h1d_tra_fix() if x_y_z_rx_ry_rz[0]=="1" else (h1d_tra_free() if stiffx==0.0 else h1d_tra_flex(stiffx))
    _y  = h1d_tra_fix() if x_y_z_rx_ry_rz[1]=="1" else (h1d_tra_free() if stiffy==0.0 else h1d_tra_flex(stiffy))
    _z  = h1d_tra_fix() if x_y_z_rx_ry_rz[2]=="1" else (h1d_tra_free() if stiffz==0.0 else h1d_tra_flex(stiffz))
    _rx = h1d_rot_fix() if x_y_z_rx_ry_rz[3]=="1" else (h1d_rot_free() if stiffrx==0.0 else h1d_rot_flex(stiffrx))
    _ry = h1d_rot_fix() if x_y_z_rx_ry_rz[4]=="1" else (h1d_rot_free() if stiffry==0.0 else h1d_rot_flex(stiffry))
    _rz = h1d_rot_fix() if x_y_z_rx_ry_rz[5]=="1" else (h1d_rot_free() if stiffrz==0.0 else h1d_rot_flex(stiffrz))
    return _x, _y, _z, _rx, _ry, _rz


    
def add_RelConnectsStructuralMember(name, beamname, pos, x_y_z_rx_ry_rz="111100", stiffx=0.0, stiffy=0.0, stiffz=0.0, stiffrx=0.0, stiffry=0.0, stiffrz=0.0):
    """  add hinge1D
    With:
      - name = name of hinge
      - beamname = name of beam
      - pos = beg, end, both
      - x_y_z_rx_ry_rz = where 1 means fixed, 0 means free, 0 + stiff-value means flexible
      - stiff... = stiffnes in MeganewtonPerMeter (for translation) or MeganewtonMeterPerRadian (for rotation)
    """
    _beam = getAdmObj(beamname, _type = "StructuralCurveMember")
    
    pos = pos.lower()
    if pos in ['beg', 0]:
        _pos = Position.Begin
    elif pos in ['end', 1]:
        _pos = Position.End
    elif pos in ['both', 2]:
        _pos = Position.Both
    
    _hinge = RelConnectsStructuralMember(newGuid(), name, _beam)
    _x, _y, _z, _rx, _ry, _rz = h1d_get_constraints(x_y_z_rx_ry_rz, stiffx, stiffy, stiffz, stiffrx, stiffry, stiffrz)
    
    obj_set_attrs(_hinge, {"Position": _pos, "TranslationX": _x, "TranslationY": _y, "TranslationZ": _z, "RotationX": _rx, "RotationY": _ry, "RotationZ": _rz})
    admService.AddItemToModel(model, _hinge)
    return _hinge


hingelst = [("h1", "b1", "end")]
for it in hingelst:
    add_RelConnectsStructuralMember(*it)


#add support in node
def add_StructuralPointSupport(name, nodename, x_y_z_rx_ry_rz="111111",* ,stiffx=0.0, stiffy=0.0, stiffz=0.0, stiffrx=0.0, stiffry=0.0, stiffrz=0.0):
    """  add support1D
    With:
      - name = name of support
      - nodename = name of node
      - x_y_z_rx_ry_rz = where 1 means fixed, 0 means free, 0 + stiff-value means flexible
      - stiffx, stiffy, ... = stiffnes in MeganewtonPerMeter (for translation) or MeganewtonMeterPerRadian (for rotation)
    """
    _node = getAdmObj(nodename, _type = "StructuralPointConnection")
    
    _support = StructuralPointSupport(newGuid(), name, _node)
    
    _x  = h1d_tra_fix() if x_y_z_rx_ry_rz[0]=="1" else (h1d_tra_free() if stiffx==0.0 else h1d_tra_flex(stiffx))
    _y  = h1d_tra_fix() if x_y_z_rx_ry_rz[1]=="1" else (h1d_tra_free() if stiffy==0.0 else h1d_tra_flex(stiffy))
    _z  = h1d_tra_fix() if x_y_z_rx_ry_rz[2]=="1" else (h1d_tra_free() if stiffz==0.0 else h1d_tra_flex(stiffz))
    _rx = h1d_rot_fix() if x_y_z_rx_ry_rz[3]=="1" else (h1d_rot_free() if stiffrx==0.0 else h1d_rot_flex(stiffrx))
    _ry = h1d_rot_fix() if x_y_z_rx_ry_rz[4]=="1" else (h1d_rot_free() if stiffry==0.0 else h1d_rot_flex(stiffry))
    _rz = h1d_rot_fix() if x_y_z_rx_ry_rz[5]=="1" else (h1d_rot_free() if stiffrz==0.0 else h1d_rot_flex(stiffrz))
    
    obj_set_attrs(_support, {"TranslationX": _x, "TranslationY": _y, "TranslationZ": _z, "RotationX": _rx, "RotationY": _ry, "RotationZ": _rz})
    admService.AddItemToModel(model, _support)
    return _support
    
supportlst = [("su1", "n1"),
              ("su2", "n2", "111001"),
              ("su3", "n3"),
              ("su4", "n4"), ]
for it in supportlst:
    add_StructuralPointSupport(*it)


 


# In[14]:


# add load group
def add_StructuralLoadGroup(name, lg_type, lg_relation="", lg_load_nature=""):
    """ add load group
    With:
     - name = load group name
     - lg_type = perm, var, seism, acc
     - lg_relation = std, excl, together
                   = Nothing for Permanent, Exclusive/Together/Standard for Variable
     - lg_load_nature = domestic, offices, congregation, shopping, storage, vehicleSmall, vehicleLarge, roofs, snow, wind, temp, constructionLoads, other
    """
    lg_type = lg_type.lower()
    if   lg_type in ["perm", "permanent"]:
        _loadgrouptype = LoadGroupType.Permanent
    elif lg_type in ["var", "variable"]:
        _loadgrouptype = LoadGroupType.Variable
    elif lg_type in ["seism", "seismic"]:
        _loadgrouptype = LoadGroupType.Seismic
    elif lg_type in ["acc", "accidental"]:
        _loadgrouptype = LoadGroupType.Accidental
    else:
        _loadgrouptype = LoadGroupType.Permanent
        
    lg_relation = lg_relation.lower()
    if   lg_relation in ["std", "standard"]:
        _relation = Relation.Standard
    elif lg_relation in ["excl", "exclusive"]:
        _relation = Relation.Exclusive
    elif lg_relation in ["togh", "together"]:
        _relation = Relation.Together
    else:
        _relation = Relation.Standard
    
    lg_load_nature = lg_load_nature.lower()
    if   lg_load_nature in ["domestic"]:
        _load = Load.Domestic
    elif lg_load_nature in ["offices"]:
        _load = Load.Offices
    elif lg_load_nature in ["congregation"]:
        _load = Load.Congregation
    elif lg_load_nature in ["shopping"]:
        _load = Load.Shopping
    elif lg_load_nature in ["storage"]:
        _load = Load.Storage
    elif lg_load_nature in ["vehiclesmall"]:
        _load = Load.VehicleSmallerThan30kN
    elif lg_load_nature in ["vehiclelarge"]:
        _load = Load.VehicleLargerThan30kN
    elif lg_load_nature in ["roofs"]:
        _load = Load.Roofs
    elif lg_load_nature in ["snow"]:
        _load = Load.Snow
    elif lg_load_nature in ["wind"]:
        _load = Load.Wind
    elif lg_load_nature in ["temperature", "temp"]:
        _load = Load.Temperature
    elif lg_load_nature in ["constructionloads", "construction"]:
        _load = Load.ConstructionLoads
    elif lg_load_nature in ["other"]:
        _load = Load.Other
    else:
        _load = Load.Domestic
    
    _lg = StructuralLoadGroup(newGuid(), name, _loadgrouptype)
    if not _loadgrouptype == LoadGroupType.Permanent:
        obj_set_attrs(_lg, {"Relation": _relation, "Load": FlexibleEnum[ Load ](_load)})
    
    admService.AddItemToModel(model, _lg)
    return _lg

lglst = [ ("lg_perm", "perm"), 
          ("lg_var1", "var", "std", "domestic"),
          ("lg_var2", "var", "std", "storage"),]
for it in lglst:
    add_StructuralLoadGroup(*it)

    


# In[15]:


# add load case
def add_StructuralLoadCase(name, action_type, lgname, lctype=None, lcduration=None):
    """ add load case
    With:
     - name = name of load case
     - action_type = sw, perm, var, acc
     - lgname = name of load group
     - lctype = sw, std, prestress, dynamic, static, temp, wind, snow, maintenance, fire, moving, seism
              = gets default-values based on action_type:  sw -> sw, perm -> std, var -> static
     - lcduration = instant, short, medium, long
                  = default: medium
    """
    _lg = getAdmObj(lgname, _type = "StructuralLoadGroup")
    
    action_type = action_type.lower()
    if   action_type in ["sw"]:
        _lc_action_type = ActionType.Permanent
        lctype = "sw"
    elif action_type in ["perm", "permanent"]:
        _lc_action_type = ActionType.Permanent
    elif action_type in ["var", "variable"]:
        _lc_action_type = ActionType.Variable
    elif action_type in ["acc", "accidental"]:
        _lc_action_type = ActionType.Accidental
    else:
        _lc_action_type = ActionType.Permanent

    if lctype is None:
        if _lc_action_type == ActionType.Permanent:
            lctype = "standard"
        elif _lc_action_type == ActionType.Variable:
            lctype = "static"
        else:
            lctype = "static"
    
    lctype = lctype.lower()
    if   lctype in ["sw", "selfweight"]:
        _lc_type = LoadCaseType.SelfWeight
    elif lctype in ["std", "standard"]:
        _lc_type = LoadCaseType.Standard
    elif lctype in ["prestress"]:
        _lc_type = LoadCaseType.Prestress
    elif lctype in ["dynamic"]:
        _lc_type = LoadCaseType.Dynamic
    elif lctype in ["static"]:
        _lc_type = LoadCaseType.Static
    elif lctype in ["temp", "temperature"]:
        _lc_type = LoadCaseType.Temperature
    elif lctype in ["wind"]:
        _lc_type = LoadCaseType.Wind
    elif lctype in ["snow"]:
        _lc_type = LoadCaseType.Snow
    elif lctype in ["maintenance"]:
        _lc_type = LoadCaseType.Maintenance
    elif lctype in ["fire"]:
        _lc_type = LoadCaseType.Fire
    elif lctype in ["moving"]:
        _lc_type = LoadCaseType.Moving
    elif lctype in ["seism", "seismic"]:
        _lc_type = LoadCaseType.Seismic
    else:
        _lc_type = LoadCaseType.Standard
    
    _lc = StructuralLoadCase(newGuid(), name, _lc_action_type, _lg, _lc_type)
    
    
    lcduration = "medium" if lcduration is None else lcduration.lower()
    if lcduration in ["instant", "instantaneous"]:
        _duration = MADM_Enum.Duration.Instantaneous
    elif lcduration in ["short"]:
        _duration = MADM_Enum.Duration.Short
    elif lcduration in ["medium"]:
        _duration = MADM_Enum.Duration.Medium
    elif lcduration in ["long"]:
        _duration = MADM_Enum.Duration.Long
    else:
        _duration = MADM_Enum.Duration.Medium

        
    if _lc_action_type == ActionType.Variable:
        obj_set_attrs(_lc, {"Duration": _duration})
    
    admService.AddItemToModel(model, _lc)
    return _lc


lclst = [ ("lc_sw"  , "sw"  , "lg_perm"),
          ("lc_perm", "perm", "lg_perm"),
          ("lc_var1", "var" , "lg_var1"),
          ("lc_var2", "var" , "lg_var2"), ]
for it in lclst:
    add_StructuralLoadCase(*it)    


# In[16]:


# add linear combination
def _combidata(lcname, factor=1.0, multiplier=1.0):
    """  helper routine for creating combination data;  loadcasename and it's factor and multiplier
    """
    return StructuralLoadCombinationData(getAdmObj(lcname), factor, multiplier)

def add_StructuralLoadCombination(name, combidatas, combi_category=None, nat_std=None, combi_type=None):
    """
    """
    _combidata_lst = System.Collections.Generic.List[ StructuralLoadCombinationData ]()
    for it in combidatas:
        _combidata_lst.Add(_combidata(*it))
    
    combi_category = "acc_en" if combi_category is None else combi_category.lower()
    if combi_category in ["natstd", "accordingnationalstandard", "acc_en", "acc_ibc"]:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.AccordingNationalStandard
    elif combi_category in ["uls", "ultimatelimitstate"]:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.UltimateLimitState
    elif combi_category in ["sls", "serviceabilitylimitstate"]:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.ServiceabilityLimitState
    elif combi_category in ["acc", "accidentallimitstate"]:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.AccidentalLimitState
    elif combi_category in ["notdefined"]:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.NotDefined
    else:
        _loadcasecombinationcategory = LoadCaseCombinationCategory.AccordingNationalStandard
    
    nat_std = "enuls" if nat_std is None else nat_std.lower()
    if   nat_std in ["enulsb", "enulssetb", "setb", "enuls"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnUlsSetB
    elif nat_std in ["enulsc", "enulssetc", "setc"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnUlsSetC
    elif nat_std in ["enacc1", "enaccidental1"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnAccidental1
    elif nat_std in ["enacc2", "enaccidental2"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnAccidental2
    elif nat_std in ["slschar", "enslschar", "enslscharacteristic", "ensls"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnSlsCharacteristic
    elif nat_std in ["slsfreq", "enslsfreq", "enslsfrequent"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnSlsFrequent
    elif nat_std in ["slsqp", "enslsqp", "enslsquasipermanent"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnSlsQuasiPermanent
    elif nat_std in ["ibclrfdult", "ibclrfdultimate"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.IbcLrfdUltimate
    elif nat_std in ["ibclrfdseism", "ibclrfdseismic"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.IbcLrfdSeismic
    elif nat_std in ["ibcasdult", "ibcasdultimate"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.IbcAsdUltimate
    elif nat_std in ["ibcasdseism", "ibcasdseismic"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.IbcAsdSeismic
    elif nat_std in ["ibcasdsls", "ibcasdserviceability"]:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.IbcAsdServiceability
    else:
        _loadcasecombinationstandard = LoadCaseCombinationStandard.EnUlsSetB
        
    combi_type = "lin" if combi_type is None else combi_type.lower()
    if combi_type in ["lin", "linear"]:
        _loadcasecombinationtype = LoadCaseCombinationType.Linear
    elif combi_type in ["nonlin", "nonlinear", "nl"]:
        _loadcasecombinationtype = LoadCaseCombinationType.NonLinear
    elif combi_type in ["env", "envelope"]:
        _loadcasecombinationtype = LoadCaseCombinationType.Envelope
    else:
        _loadcasecombinationtype = LoadCaseCombinationType.Linear
    
    _combi = StructuralLoadCombination(newGuid(), name, _loadcasecombinationcategory, _combidata_lst)
    
    if _loadcasecombinationcategory == LoadCaseCombinationCategory.AccordingNationalStandard:
        _attrs = {"Description": "", "NationalStandard": _loadcasecombinationstandard, "Type": _loadcasecombinationtype}
    else:
        _attrs = {"Description": "", "Type": _loadcasecombinationtype}
    obj_set_attrs(_combi, _attrs)
    
    admService.AddItemToModel(model, _combi)
    return _combi


combilst = [ ("c_enulsb", [ ("lc_sw", 1.00), ("lc_perm", 1.00), ("lc_var1", 1.0), ("lc_var2", 1.0) ], "acc_EN", "EnUlsSetB"),
             ("c_lin1",   [ ("lc_sw", 1.35), ("lc_perm", 1.35), ("lc_var1", 1.5), ("lc_var2", 1.5) ], "ULS" )]
for it in combilst:
    add_StructuralLoadCombination(*it)    


# In[17]:


# add loads   in a point
def add_StructuralPointAction_OnNode(name, value_kN, direction, lc_name, node_name):
    """
    """
    _forcevalue_kN = Force(value_kN, ForceUnit.Kilonewton)
    _lc = getAdmObj(lc_name, _type="StructuralLoadCase")
    
    _node = getAdmObj(node_name, _type="StructuralPointConnection")
    _tpointref = PointStructuralReferenceOnPoint(_node)
    
    _pointload = StructuralPointAction[PointStructuralReferenceOnPoint](newGuid(), name, _forcevalue_kN, _lc, PointForceAction.InNode, _tpointref)
    
    direction = direction.lower()
    if   direction in ["x"]:
        _dir = ActionDirection.X
    elif direction in ["y"]:
        _dir = ActionDirection.Y
    elif direction in ["z"]:
        _dir = ActionDirection.Z
    else:
        _dir = ActionDirection.Z
        
    obj_set_attrs(_pointload, {"Direction": _dir, "CoordinateSystem": CoordinateSystem.Global})
    
    admService.AddItemToModel(model, _pointload)
    return _pointload



add_StructuralPointAction_OnNode("pl1", 5, "X", "lc_perm", "n5")


def add_StructuralPointAction_OnBeam(name, value_kN, direction, lc_name, beam_name, lcs_type=None, coord_def=None, position=0.0, origin=None):
    """
    """
    _forcevalue_kN = Force(value_kN, ForceUnit.Kilonewton)
    _lc = getAdmObj(lc_name, _type="StructuralLoadCase")
    
    _beam = getAdmObj(beam_name, _type="StructuralCurveMember")
    _tpointref = PointStructuralReferenceOnBeam(_beam)
    
    
    attrs = {}
    lcs_type = "GCS" if lcs_type is None else lcs_type
    lcs_type = lcs_type.lower()
    if   lcs_type in ["gcs", "global"]:
        _lcs_type = CoordinateSystem.Global
    elif lcs_type in ["lcs", "local"]:
        _lcs_type = CoordinateSystem.Global
    else:
        _lcs_type = CoordinateSystem.Global
    attrs["CoordinateSystem"] = _lcs_type
    
    origin = "FromStart" if origin is None else origin
    origin = origin.lower()
    if origin in ["fromstart", "start", "begin"]:
        _origin = Origin.FromStart
    elif origin in ["fromend", "end"]:
        _origin = Origin.FromEnd
    else:
        _origin = Origin.FromStart
    attrs["Origin"] = _origin
    
    coord_def = "abs" if coord_def is None else coord_def
    coord_def = coord_def.lower()
    if   coord_def in ["abs"]:
        _coord_def = CoordinateDefinition.Absolute
    elif coord_def in ["rel"]:
        _coord_def = CoordinateDefinition.Relative
    else:
        _coord_def = CoordinateDefinition.Absolute
    attrs["CoordinateDefinition"] = _coord_def
    
    if _coord_def == CoordinateDefinition.Absolute:
        attrs["AbsolutePositionX"] = Length(position, LengthUnit.Meter)
    else:
        attrs["RelativePositionX"] = position
    
    direction = direction.lower()
    if   direction in ["x"]:
        _dir = ActionDirection.X
    elif direction in ["y"]:
        _dir = ActionDirection.Y
    elif direction in ["z"]:
        _dir = ActionDirection.Z
    else:
        _dir = ActionDirection.Z
    attrs["Direction "] = _dir
    
    _pointload = StructuralPointAction[PointStructuralReferenceOnBeam](newGuid(), name, _forcevalue_kN, _lc, PointForceAction.OnBeam, _tpointref)
    obj_set_attrs(_pointload, attrs)
    
    admService.AddItemToModel(model, _pointload)
    return _pointload


add_StructuralPointAction_OnBeam("pl2", 5, "X", "lc_perm", "b5", coord_def="Rel", position = 0.8)


# In[ ]:





# In[ ]:





# In[ ]:





# In[18]:


def write_ADM_to_SAF():
    try:
        print("Validation of model: ", end="")
        admService.ValidateModel(model)
        print("OK")
    except ValidationException as err:
        print("FAILED")
        print("Validation Exception")
        print(f"Exception = \n{err}")
        return
    except Exception as err:
        print("FAILED")
        print(f"Exception Type = \n -> {type(err)}")
        print(f"Exception = \n{err}")        
        return

    try:
        excelBootstrapper = ExcelModuleBootstrapper()
        excelScope = excelBootstrapper.CreateThreadedScope()
        excelWriterService = excelScope.GetService[IExcelWriter]()
        print("Creation of SAF file: ", end="")
        fStream = None
        fStream = System.IO.File.Create(safxlsx_fname)
        print("OK")
    except System.IO.IOException as err:
        print("FAILED")
        print("Error while creating the file stream")
        print(f"Exception Message = \n{err}")
    except Exception as err:
        print("FAILED")
        print(f"Exception Type = \n -> {type(err)}")
        print(f"Exception Message = \n{err}")
    else:
        print(f"Writing to: \n   -> {safxlsx_fname}")
        ver = System.Version(2, 1, 0)
        excelWriterService.Write(fStream, model, ver)
    finally:
        if fStream:
            fStream.Flush()
            fStream.Close()
        print("Done.")
        
        
write_ADM_to_SAF()


# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:





# In[ ]:




